{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module Syntax.Pretty
    ( module Syntax.Pretty
    , module Text.PrettyPrint
    ) where

import Control.Monad
import Control.Monad.State
import Data.List
import Data.IORef
import System.IO.Unsafe ( unsafePerformIO )
import Text.PrettyPrint

import Syntax.Internal
import Syntax.Position

import {-# SOURCE #-} TypeChecker.UnfoldConstants -- REMOVE_IF_HADDOCK

import Utils.Names	    ( noNamesUsed, fresh )
import Utils.Monads
import Utils.Pairs

{- REMOVE_IF_HADDOCK
-- | Hack to make Haddock accept cyclic module dependencies.
--   The real function is 'TypeChecker.Computation.unfoldConstants'.
unfoldConstants :: Value -> Value

-- | Hack to make Haddock accept cyclic module dependencies.
--   The real function is 'TypeChecker.Values.projectionNames'.
projectionNames :: Value -> (Value, [String])

REMOVE_IF_HADDOCK -}

-- The Pretty class -------------------------------------------------------

class Pretty a where
    pretty	:: a -> Doc
    prettyPrec	:: Int -> a -> Doc

    pretty	    = prettyPrec 0
    prettyPrec _    = pretty

-- Show instances ---------------------------------------------------------

instance Show (AnnotatedTerm t v d)	    where   show = show . pretty
instance Show Value			    where   show = show . pretty
instance Show SortValue			    where   show = show . pretty
instance Show Program			    where   show = show . pretty
instance Show (AnnotatedDefinition t v d)   where   show = show . pretty

instance Pretty Doc where
    pretty = id

instance Pretty a => Pretty [a] where
    pretty = brackets . fsep . punctuate comma . map pretty

-- Pretty ranges

instance Pretty Range where
    pretty = text . show

-- Pretty programs --------------------------------------------------------

instance Pretty Program where
    pretty prog = vcat $ map pretty prog

-- Pretty definitions -----------------------------------------------------

instance Pretty (AnnotatedDefinition t v d) where
    pretty (NormalDef name mt c) =
	vcat $ maybe empty (\t ->
		sep [ prettyName name <+> text "::"
		    , nest 2 $ pretty t
		    ]
		) mt : [text name <+> text "=" <+> pretty c]
    pretty (NewAxiom name t) =
	sep [ prettyName name <+> text "::"
	    , nest 2 $ pretty t
	    ]

instance Pretty (AnnotatedLocalDefinition t v d) where
    pretty (LocalDef name mt e _) =
	vcat	[ maybe empty (\t ->
		    sep [ prettyName name <+> text "::"
			, nest 2 $ pretty t
			]
		    ) mt
		, sep	[ prettyName name
			, nest 2 $ text "=" <+> pretty e
			]
		]

-- Pretty terms -----------------------------------------------------------

data PrettyTerm	t v d
	= PApp (AnnotatedTerm t v d) (AnnotatedTerm t v d)
	| PName Name
	| PLam [Name] (AnnotatedTerm t v d)
	| PStar Universe
	| PPi [Name] (AnnotatedTerm t v d) (AnnotatedTerm t v d)
	| PFun (AnnotatedTerm t v d) (AnnotatedTerm t v d)
	| PPlugin Name [Term]
	| PLet [AnnotatedLocalDefinition t v d] (AnnotatedTerm t v d)

prettyView :: AnnotatedTerm t v d -> PrettyTerm t v d
prettyView e =
    case e of
	App e1 e2   -> PApp e1 e2
	Var x _   -> PName x
	Def x _   -> PName x
	Lam p e _ -> case prettyView e of
			    PLam xs e'	-> PLam (p:xs) e'
			    _		-> PLam [p] e
	Star u	-> PStar u
	Prop		-> PName "Prop"
	Pi "_" t e _	-> PFun t e
	Pi p t e _	->
		case prettyView e of
		    PPi ps t' e' | t == t'  -> PPi (p:ps) t e'
		    _			    -> PPi [p] t e
	CallPlugin x es _	-> PPlugin x es
	Let ds e		-> PLet ds e

inf = maxBound :: Int
appPrec = inf - 2
projPrec = inf - 1

instance Pretty (AnnotatedTerm t v d) where

    prettyPrec p e =
	case prettyView e of
	    PApp e1 e2	-> 
		mparens (p > appPrec) $
		    sep [ prettyPrec appPrec e1
			, nest 2 $ prettyPrec (appPrec + 1) e2
			]

	    PName x	-> prettyName x

	    PLam ps e	->
		mparens (p>0) $
		    sep	[ sep [ text "\\" <> fsep (map text ps), text "->" ]
			, nest 2 $ pretty e
			]

	    PStar 0 -> text "Set"
	    PStar 1 -> text "Type"
	    PStar u -> text "#" <> pretty u

	    PFun a b		->
		mparens (p>0) $
		    sep [ prettyPrec 1 a <+> text "->"
			, pretty b
			]
	    PPi ps a b		->
		mparens (p>0) $
		    sep [ parens (fsep $
				    (punctuate (text ",") $ map text ps)
				    ++	[ text "::"
					, nest 2 $ pretty a
					]
				    )
			    <+> text "->"
			, pretty b
			]

	    PPlugin x es ->
		sep [ text $ x ++ "-plugin"
		    , nest 2 $ parens	$ fsep
					$ punctuate (text ",")
					$ map pretty es
		    ]

	    PLet ds e		->
		mparens (p>0) $
		    fsep    [ text "let" <+> vcat (map pretty ds)
			    , text "in" <+> pretty e
			    ]

prettyName x = mparens (isOperator x) $ text x

-- Pretty values ----------------------------------------------------------

instance Pretty Value where
    prettyPrec p v = flip evalState noNamesUsed $ prettyM p v
	where
	    prettyM p (VId x t)	    = return $ prettyName x
	    prettyM p (VDef x d)    = return $ prettyName x
	    prettyM p (VSort s)	    = return $ pretty s
	    prettyM p (VApp v w)    =
		do  dv <- prettyM 9 v
		    dw <- prettyM 10 w
		    return $ mparens (p>9) $
			sep [ dv, nest 2 dw ]

	    prettyM p v@(VPi _ _ _) =
		do  ds <- pPi v
		    return $ mparens (p>0) $
			fsep $ punctuate (text " ->") ds
		where
		    pPi (VPi v f t)
			| clName f == dontCare =
			    do	dv <- prettyM 1 v
				ds <- pPi $ clApp f $ VId dontCare v
				return (dv:ds)
			| otherwise =
			    do  dv  <- prettyM 0 v
				x   <- fresh $ clName f
				ds  <- pPi $ clApp f $ VId x v
				let dv' = parens $ sep  [ text x <+> text "::"
							, nest 2 dv
							]
				return (dv':ds)
		    pPi v = (:[]) <$> prettyM 0 v

	    prettyM p v@(VLam _ _) =
		do  (xs, df) <- pLam v
		    return $ mparens (p>0) $
			sep [ text "\\" <> fsep (map text $ xs ++ ["->"])
			    , nest 2 df
			    ]
		where
		    pLam (VLam f t) =
			do  x <- fresh $ clName f
			    let VPi v _ _ = unfoldConstants t
			    (xs,df) <- pLam $ clApp f $ VId x v
			    return (x:xs, df)
		    pLam v =
			do  dv <- prettyM 0 v
			    return ([], dv)
	    
	    prettyM p (VTop _)	= return $ text "Top"
	    prettyM p (VTag v)	= do d <- prettyM appPrec v
				     return $ braces d

instance Pretty SortValue where
    pretty (VSUniv u)	= text "#" <> pretty u
    pretty VSProp	= text "Prop"

-- Pretty ints ------------------------------------------------------------

instance Pretty Int where
    pretty = text . show

-- Helpers ----------------------------------------------------------------

mparens True	= parens
mparens _	= id

