{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

{-| The FOL plugin.
-}

module Plugins.FOL where -- (plugin) where

import Control.Monad
import Control.Monad.Trans	    ( liftIO )

import Syntax.Position		    ( Range )
import Syntax.Internal		    ( Term )
import Syntax.Values		    ( Value, Type, getSort, unEl, vTop )

import TypeChecker		    ( inferType )
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Context
import TypeChecker.Values.DeBruijn  ( unBruijn )
import TypeChecker.Values.Refresh   ( refresh )
import TypeChecker.MetaVars

import Utils.Monads

import Plugins
import Plugins.FOL.Types	    ( FOLGoal(..), Goal(..) )
import Plugins.FOL.Errors	    ( noInferenceForFOLError )
import Plugins.FOL.Options
import Plugins.FOL.Translation
import Plugins.FOL.Solve	    ( solveGoals )
import Plugins.FOL.Errors
import Plugins.FOL.Arities
import Plugins.FOL.Arguments

-- The FOL plugin ---------------------------------------------------------

plugin :: SomePlugin
plugin = mkPlugin thePlugin

thePlugin =
    Plugin  { plName		= "fol"
            , plHelp            = helpFOL
	    , plInit		= \_ -> return []
	    , plFinal		= finalFOL
	    , plCheck		= checkFOL
	    , plInfer		= \_ _ _ _ -> failure noInferenceForFOLError
	    , plOptions		= simpleOptions options
	    , plDefaultOptions	= defaultOptions
	    }

finalFOL :: FOLOptions -> [Goal] -> TC ()
finalFOL opts goals =
    do	fol_goals <- mapM translateGoal goals
	solveGoals opts fol_goals
    where
	translateGoal (Goal r args gamma vts t) =
	    sayWhere r $
	    sayWhen whenTranslatingToFOL $
	    do	gamma' <- refresh gamma
		(t',vts') <- inContext gamma' $ unBruijn =<< refresh (t,vts)
		((pre,goal),lambdas) <-
		    runM args $ (,) <$> mapM (uncurry argumentToFOL) vts'
				    <*> valueToFOL (unEl t')
		let (lambdas',pre',goal')
			| optBinaryApp opts = (lambdas,pre,goal)
			| otherwise	    = removeBinaryApps (lambdas,pre,goal)
		return $ FOLGoal r args (lambdas', pre') goal' (map fst vts') (unEl t')


checkFOL :: FOLOptions -> [Goal] -> Range -> [Term] -> Type ->
	    TC (Value, [Goal])
checkFOL = checkFOL' NoIncludes

checkFOL' with_includes opts goals r es v =
    do	unless (isProp v) $ failure $ notOfTypePropError v (getSort v)
	let args = parseArguments with_includes es
	vs <- mapM inferArgument $ premisses args
	gamma <- getContext
	return (vTop, Goal r args gamma vs v : goals)

helpFOL :: String
helpFOL = unlines
  [ "Invokes a automated first-order logic (FOL) prover."
  , "The verbosity levels are interpreted as follows: "
  , "    >= 2: Print goals sent to FOL provers."
  , "    >= 3: Print goals in FOL form and proofs."
  , "    >= 4: Print stderr output received from FOL prover."
  , "    >= 5: Print full FOL prover input and output."
  ]
